# Copyright 2020 Silicon Compiler Authors. All Rights Reserved.

import json
import os
import requests
import shutil
import time
import urllib.parse
import tarfile
import tempfile
import multiprocessing

from siliconcompiler import utils, SiliconCompilerError
from siliconcompiler import NodeStatus as SCNodeStatus
from siliconcompiler._metadata import default_server
from siliconcompiler.flowgraph import nodes_to_execute
from siliconcompiler.remote import JobStatus

# Step name to use while logging
remote_step_name = 'remote'

# Client / server timeout
__timeout = 10

# Generate warning if no server is configured
__warn_if_no_server = True

# Multiprocessing interface.
multiprocessor = multiprocessing.get_context('spawn')

__tos_str = '''Please review the SiliconCompiler cloud's terms of service:

https://www.siliconcompiler.com/terms

In particular, please ensure that you have the right to distribute any IP
which is contained in designs that you upload to the service. This public
service, provided by SiliconCompiler, is not intended to process proprietary IP.
'''


###################################
def get_base_url(chip):
    '''Helper method to get the root URL for API calls, given a Chip object.
    '''

    rcfg = get_remote_config(chip, False)
    remote_host = rcfg['address']
    if 'port' in rcfg:
        remote_port = rcfg['port']
    else:
        remote_port = 443
    remote_host += ':' + str(remote_port)
    if remote_host.startswith('http'):
        remote_protocol = ''
    else:
        remote_protocol = 'https://' if str(remote_port) == '443' else 'http://'
    return remote_protocol + remote_host


def get_remote_manifest(chip):
    return f'{chip.getworkdir()}/sc_remote.pkg.json'


###################################
def __post(chip, url, post_action, success_action, error_action=None):
    '''
    Helper function to handle the post request
    '''
    redirect_url = urllib.parse.urljoin(get_base_url(chip), url)

    timeouts = 0
    while redirect_url:
        try:
            resp = post_action(redirect_url)
        except requests.Timeout:
            timeouts += 1
            if timeouts > 10:
                raise SiliconCompilerError('Server communications timed out', chip=chip)
            time.sleep(10)
            continue
        except Exception as e:
            raise SiliconCompilerError(f'Server communications error: {e}', chip=chip)

        code = resp.status_code
        if 200 <= code and code < 300:
            return success_action(resp)

        try:
            msg_json = resp.json()
            if 'message' in msg_json:
                msg = msg_json['message']
            else:
                msg = resp.text
        except requests.JSONDecodeError:
            msg = resp.text

        if 300 <= code and code < 400:
            if 'Location' in resp.headers:
                redirect_url = resp.headers['Location']
                continue

        if error_action:
            return error_action(code, msg)
        else:
            raise SiliconCompilerError(f'Server responded with {code}: {msg}', chip=chip)


###################################
def __build_post_params(chip, verbose, job_name=None, job_hash=None):
    '''
    Helper function to build the params for the post request
    '''
    # Use authentication if necessary.
    post_params = {}

    if job_hash:
        post_params['job_hash'] = job_hash

    if job_name:
        post_params['job_id'] = job_name

    rcfg = get_remote_config(chip, verbose)

    if ('username' in rcfg) and ('password' in rcfg) and \
       (rcfg['username']) and (rcfg['password']):
        post_params['username'] = rcfg['username']
        post_params['key'] = rcfg['password']

    return post_params


###################################
def _remote_preprocess(chip):
    '''
    Helper method to run a local import stage for remote jobs.
    '''

    # Ensure packages with python sources are copied
    for key in chip.allkeys():
        key_type = chip.get(*key, field='type')

        if 'dir' in key_type or 'file' in key_type:
            for _, step, index in chip.schema._getvals(*key, return_defvalue=False):
                packages = chip.get(*key, field='package', step=step, index=index)
                force_copy = False
                for package in packages:
                    if not package:
                        continue
                    if package.startswith('python://'):
                        force_copy = True
                if force_copy:
                    chip.set(*key, True, field='copy', step=step, index=index)

    # Collect inputs into a collection directory only for remote runs, since
    # we need to send inputs up to the server.
    cfg = get_remote_config(chip, False)
    chip.collect(whitelist=cfg.setdefault('directory_whitelist', []))


###################################
def _log_truncated_stats(chip, status, nodes_with_status, nodes_to_print):
    '''
    Helper method to log truncated information about flowgraph nodes
    with a given status, on a single line.
    Used to print info about all statuses besides NodeStatus.RUNNING.
    '''

    num_nodes = len(nodes_with_status)
    if num_nodes > 0:
        nodes_log = f'  {status.title()} ({num_nodes}): '
        log_nodes = []
        for i in range(min(nodes_to_print, num_nodes)):
            log_nodes.append(nodes_with_status[i][0])
        if num_nodes > nodes_to_print:
            log_nodes.append('...')
        nodes_log += ', '.join(log_nodes)
        chip.logger.info(nodes_log)


###################################
def _process_progress_info(chip, progress_info, recorded_nodes, all_nodes, nodes_to_print=3):
    '''
    Helper method to log information about a remote run's progress,
    based on information returned from a 'check_progress/' call.
    '''

    completed = []
    try:
        # Decode response JSON, if possible.
        job_info = json.loads(progress_info['message'])

        # Sort and store info about the job's progress.
        chip.logger.info("Job is still running. Status:")

        nodes_to_log = {}
        for node, node_info in job_info.items():
            status = node_info['status']
            nodes_to_log.setdefault(status, []).append((node, node_info))

            if SCNodeStatus.is_done(status):
                # collect completed
                completed.append(node)

            if node in all_nodes:
                step, index = all_nodes[node]
                if (step, index) not in recorded_nodes:
                    chip.set('record', 'status', status, step=step, index=index)

        nodes_to_log = {key: nodes_to_log[key] for key in sorted(nodes_to_log.keys())}

        # Log information about the job's progress.
        # To avoid clutter, only log up to N completed/pending nodes, on a single line.
        # Completed, failed, and timed-out flowgraph nodes:
        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_done(stat):
                _log_truncated_stats(chip, stat, nodes, nodes_to_print)

        # Running / in-progress flowgraph nodes should all be printed:
        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_running(stat):
                chip.logger.info(f'  {stat.title()} ({len(nodes)}):')
                for node, node_info in nodes:
                    running_log = f"    {node}"
                    if 'elapsed_time' in node_info:
                        running_log += f" ({node_info['elapsed_time']})"
                    chip.logger.info(running_log)

        # Queued and pending flowgraph nodes:
        for stat, nodes in nodes_to_log.items():
            if SCNodeStatus.is_waiting(stat):
                _log_truncated_stats(chip, stat, nodes, nodes_to_print)
    except json.JSONDecodeError:
        chip.logger.info("Job is still running")
    except Exception as e:
        chip.logger.info(f"Job is still running: local error: {e}")

    return completed


def get_remote_config_file(chip, fail=True):
    if chip.get('option', 'credentials'):
        # Use the provided remote credentials file.
        cfg_file = os.path.abspath(chip.get('option', 'credentials'))

        if fail and not os.path.isfile(cfg_file):
            # Check if it's a file since its been requested by the user
            raise SiliconCompilerError(
                f'Unable to find the credentials file: {cfg_file}',
                chip=chip)
    else:
        # Use the default config file path.
        cfg_file = utils.default_credentials_file()

    return cfg_file


def get_remote_config(chip, verbose):
    '''
    Returns the remote credentials
    '''
    cfg_file = get_remote_config_file(chip)

    remote_cfg = {}
    cfg_dir = os.path.dirname(cfg_file)
    if os.path.isdir(cfg_dir) and os.path.isfile(cfg_file):
        if verbose:
            chip.logger.info(f'Using credentials: {cfg_file}')
        with open(cfg_file, 'r') as cfgf:
            remote_cfg = json.loads(cfgf.read())
    else:
        global __warn_if_no_server
        if __warn_if_no_server:
            if verbose:
                chip.logger.warning('Could not find remote server configuration: '
                                    f'defaulting to {default_server}')
            __warn_if_no_server = False
        remote_cfg = {
            "address": default_server,
            "directory_whitelist": []
        }
    if 'address' not in remote_cfg:
        raise SiliconCompilerError(
            'Improperly formatted remote server configuration - '
            'please run "sc-remote -configure" and enter your server address and '
            'credentials.', chip=chip)

    return remote_cfg


def remote_process(chip):
    '''
    Dispatch the Chip to a remote server for processing.
    '''
    should_resume = not chip.get('option', 'clean')
    remote_resume = should_resume and chip.get('record', 'remoteid')

    # Pre-process: Run an starting nodes locally, and upload the
    # in-progress build directory to the remote server.
    # Data is encrypted if user / key were specified.
    # run remote process
    if chip.get('arg', 'step'):
        raise SiliconCompilerError('Cannot pass [arg,step] parameter into remote flow.', chip=chip)
    if chip.get('arg', 'index'):
        raise SiliconCompilerError('Cannot pass [arg,index] parameter into remote flow.', chip=chip)
    # Only run the pre-process step if the job doesn't already have a remote ID.
    if not remote_resume:
        _remote_preprocess(chip)

    # Run the job on the remote server, and wait for it to finish.
    # Set logger to indicate remote run
    chip._init_logger(step=remote_step_name, index=None, in_run=True)
    _remote_run(chip)

    # Restore logger
    chip._init_logger(in_run=True)


###################################
def _remote_run(chip):
    '''
    Helper method to run a job stage on a remote compute cluster.
    Note that files will not be copied to the remote stage; typically
    the source files will be copied into the cluster's storage before
    calling this method.
    If the "-remote" parameter was not passed in, this method
    will print a warning and do nothing.
    This method assumes that the given stage should not be skipped,
    because it is called from within the `Chip.run(...)` method.

    '''

    # Ask the remote server to start processing the requested step.
    check_interval = _request_remote_run(chip)

    # Remove the local 'import.tar.gz' archive.
    local_archive = os.path.join(chip.getworkdir(),
                                 'import.tar.gz')
    if os.path.isfile(local_archive):
        os.remove(local_archive)

    # Run the main 'check_progress' loop to monitor job status until it finishes.
    remote_run_loop(chip, check_interval)


###################################
def remote_run_loop(chip, check_interval):
    # Wrapper to allow for capturing of Ctrl+C
    try:
        __remote_run_loop(chip, check_interval)
    except KeyboardInterrupt:
        manifest_path = get_remote_manifest(chip)
        reconnect_cmd = f'sc-remote -cfg {manifest_path} -reconnect'
        cancel_cmd = f'sc-remote -cfg {manifest_path} -cancel'
        chip.logger.info('Disconnecting from remote job')
        chip.logger.info(f'To reconnect to this job use: {reconnect_cmd}')
        chip.logger.info(f'To cancel this job use: {cancel_cmd}')
        raise SiliconCompilerError('Job canceled by user keyboard interrupt')


###################################
def __remote_run_loop(chip, check_interval):
    # Check the job's progress periodically until it finishes.
    is_busy = True
    all_nodes = {}
    completed = []
    result_procs = []

    recorded = []

    for step, index in nodes_to_execute(chip):
        if SCNodeStatus.is_done(chip.get('record', 'status', step=step, index=index)):
            continue
        all_nodes[f'{step}{index}'] = (step, index)

    def import_manifests():
        changed = False
        for step, index in all_nodes.values():
            if (step, index) in recorded:
                continue

            manifest = os.path.join(chip.getworkdir(step=step, index=index),
                                    'outputs',
                                    f'{chip.design}.pkg.json')
            if os.path.exists(manifest):
                try:
                    chip.schema.read_journal(manifest)
                    recorded.append((step, index))
                    changed = True
                except:  # noqa E722
                    # Import may fail if file is still getting written
                    pass
        return changed

    def schedule_download(node):
        node_proc = multiprocessor.Process(target=fetch_results,
                                           args=(chip, node))
        node_proc.start()
        result_procs.append(node_proc)
        if node is None:
            node = 'final result'
        chip.logger.info(f'    {node}')

    while is_busy:
        time.sleep(check_interval)
        import_manifests()
        new_completed, is_busy = check_progress(chip, recorded, all_nodes)

        if chip._dash:
            chip._dash.update_manifest()

        nodes_to_fetch = []
        for node in new_completed:
            if node not in completed:
                nodes_to_fetch.append(node)
                completed.append(node)
        if nodes_to_fetch:
            chip.logger.info('  Fetching completed results:')
            for node in nodes_to_fetch:
                schedule_download(node)

    # Done: try to fetch any node results which still haven't been retrieved.
    chip.logger.info('Remote job completed! Retrieving final results...')
    for node_name in all_nodes:
        if node_name not in completed:
            schedule_download(node_name)
    schedule_download(None)

    # Make sure all results are fetched before letting the client issue
    # a deletion request.
    for proc in result_procs:
        proc.join()

    # Read in node manifests
    import_manifests()

    # Un-set the 'remote' option to avoid from/to-based summary/show errors
    chip.unset('option', 'remote')

    if chip._dash:
        chip._dash.update_manifest()


###################################
def check_progress(chip, recorded_nodes, all_nodes):
    try:
        is_busy_info = is_job_busy(chip)
        is_busy = is_busy_info['busy']
        completed = []
        if is_busy:
            completed = _process_progress_info(chip,
                                               is_busy_info,
                                               recorded_nodes,
                                               all_nodes)
        return completed, is_busy
    except Exception as e:
        # Sometimes an exception is raised if the request library cannot
        # reach the server due to a transient network issue.
        # Retrying ensures that jobs don't break off when the connection drops.
        chip.logger.info(f"Unknown network error encountered: retrying: {e}")
        return [], True


###################################
def _request_remote_run(chip):
    '''
    Helper method to make a web request to start a job stage.
    '''

    remote_resume = not chip.get('option', 'clean') and chip.get('record', 'remoteid')
    # Only package and upload the entry steps if starting a new job.
    if not remote_resume:
        upload_file = tempfile.TemporaryFile(prefix='sc', suffix='remote.tar.gz')
        with tarfile.open(fileobj=upload_file, mode='w:gz') as tar:
            tar.add(chip.getworkdir(), arcname='')
        # Flush file to ensure everything is written
        upload_file.flush()

    remote_status = _remote_ping(chip)

    if remote_status['status'] != 'ready':
        raise SiliconCompilerError('Remote server is not available.', chip=chip)

    __print_tos(chip, remote_status)

    if 'pre_upload' in remote_status:
        chip.logger.info(remote_status['pre_upload']['message'])
        time.sleep(remote_status['pre_upload']['delay'])

    # Make the actual request, streaming the bulk data as a multipart file.
    # Redirected POST requests are translated to GETs. This is actually
    # part of the HTTP spec, so we need to manually follow the trail.
    post_params = {
        'chip_cfg': chip.schema.cfg,
        'params': __build_post_params(chip,
                                      False,
                                      job_hash=chip.get('record', 'remoteid'))
    }

    post_files = {'params': json.dumps(post_params)}
    if not remote_resume:
        post_files['import'] = upload_file
        upload_file.seek(0)

    def post_action(url):
        return requests.post(url,
                             files=post_files,
                             timeout=__timeout)

    def success_action(resp):
        return resp.json()

    resp = __post(chip, '/remote_run/', post_action, success_action)
    if not remote_resume:
        upload_file.close()

    if 'message' in resp and resp['message']:
        chip.logger.info(resp['message'])
    chip.set('record', 'remoteid', resp['job_hash'])

    manifest = get_remote_manifest(chip)
    chip.write_manifest(manifest)

    chip.logger.info(f"Your job's reference ID is: {resp['job_hash']}")

    return remote_status['progress_interval']


###################################
def is_job_busy(chip):
    '''
    Helper method to make an async request asking the remote server
    whether a job is busy, or ready to accept a new step.
    Returns True if the job is busy, False if not.
    '''

    # Make the request and print its response.
    def post_action(url):
        params = __build_post_params(chip,
                                     False,
                                     job_hash=chip.get('record', 'remoteid'),
                                     job_name=chip.get('option', 'jobname'))
        return requests.post(url,
                             data=json.dumps(params),
                             timeout=__timeout)

    def error_action(code, msg):
        return {
            'busy': True,
            'message': ''
        }

    def success_action(resp):
        json_response = json.loads(resp.text)

        if json_response['status'] != JobStatus.RUNNING:
            if json_response['status'] == JobStatus.REJECTED:
                chip.logger.error(f'Job was rejected: {json_response["message"]}')
            elif json_response['status'] != JobStatus.UNKNOWN:
                chip.logger.info(f'Job status: {json_response["status"]}')
        info = {
            'busy': json_response['status'] == JobStatus.RUNNING,
            'message': None
        }

        if isinstance(json_response['message'], str):
            info['message'] = json_response['message']
        else:
            info['message'] = json.dumps(json_response['message'])
        return info

    info = __post(chip,
                  '/check_progress/',
                  post_action,
                  success_action,
                  error_action=error_action)

    if not info:
        info = {
            'busy': True,
            'message': ''
        }
    return info


###################################
def cancel_job(chip):
    '''
    Helper method to request that the server cancel an ongoing job.
    '''

    def post_action(url):
        return requests.post(url,
                             data=json.dumps(__build_post_params(
                                chip,
                                False,
                                job_hash=chip.get('record', 'remoteid'))),
                             timeout=__timeout)

    def success_action(resp):
        return json.loads(resp.text)

    return __post(chip, '/cancel_job/', post_action, success_action)


###################################
def delete_job(chip):
    '''
    Helper method to delete a job from shared remote storage.
    '''

    def post_action(url):
        return requests.post(url,
                             data=json.dumps(__build_post_params(
                                chip,
                                False,
                                job_hash=chip.get('record', 'remoteid'))),
                             timeout=__timeout)

    def success_action(resp):
        return resp.text

    return __post(chip, '/delete_job/', post_action, success_action)


###################################
def fetch_results_request(chip, node, results_fd):
    '''
    Helper method to fetch job results from a remote compute cluster.
    Optional 'node' argument fetches results for only the specified
    flowgraph node (e.g. "floorplan0")

       Returns:
       * 0 if no error was encountered.
       * [response code] if the results could not be retrieved.
    '''

    # Set the request URL.
    job_hash = chip.get('record', 'remoteid')

    # Fetch results archive.
    def post_action(url):
        post_params = __build_post_params(chip, False)
        if node:
            post_params['node'] = node
        return requests.post(url,
                             data=json.dumps(post_params),
                             stream=True,
                             timeout=__timeout)

    def success_action(resp):
        shutil.copyfileobj(resp.raw, results_fd)
        return 0

    def error_action(code, msg):
        # Results are fetched in parallel, and a failure in one node
        # does not necessarily mean that the whole job failed.
        if node:
            chip.logger.warning(f'Could not fetch results for node: {node}')
        else:
            chip.logger.warning('Could not fetch results for final results.')
        return 404

    return __post(chip,
                  f'/get_results/{job_hash}.tar.gz',
                  post_action,
                  success_action,
                  error_action=error_action)


###################################
def fetch_results(chip, node):
    '''
    Helper method to fetch and open job results from a remote compute cluster.
    Optional 'node' argument fetches results for only the specified
    flowgraph node (e.g. "floorplan0")
    '''

    # Collect local values.
    job_hash = chip.get('record', 'remoteid')
    local_dir = chip.get('option', 'builddir')

    # Set default results archive path if necessary, and fetch it.
    with tempfile.TemporaryDirectory(prefix=f'sc_{job_hash}_', suffix=f'_{node}') as tmpdir:
        results_path = os.path.join(tmpdir, 'result.tar.gz')

        with open(results_path, 'wb') as rd:
            results_code = fetch_results_request(chip, node, rd)

        # Note: the server should eventually delete the results as they age out (~8h), but this will
        # give us a brief period to look at failed results.
        if results_code:
            return

        # Unzip the results.
        # Unauthenticated jobs get a gzip archive, authenticated jobs get nested archives.
        # So we need to extract and delete those.
        # Archive contents: server-side build directory. Format:
        # [job_hash]/[design]/[job_name]/[step]/[index]/...
        try:
            with tarfile.open(results_path, 'r:gz') as tar:
                tar.extractall(path=tmpdir)
        except tarfile.TarError as e:
            chip.logger.error(f'Failed to extract data from {results_path}: {e}')
            return

        work_dir = os.path.join(tmpdir, job_hash)
        if os.path.exists(work_dir):
            shutil.copytree(work_dir, local_dir, dirs_exist_ok=True)
        else:
            chip.logger.error(f'Empty file returned from remote for: {node}')
            return


def _remote_ping(chip):
    # Make the request and print its response.
    rcfg = __build_post_params(chip, True)

    def post_action(url):
        return requests.post(url,
                             data=json.dumps(rcfg),
                             timeout=__timeout)

    def success_action(resp):
        return resp.json()

    response_info = __post(chip, '/check_server/', post_action, success_action)
    if not response_info:
        raise ValueError('Server response is not valid.')

    return response_info


###################################
def __print_tos(chip, response_info):
    # Print terms-of-service message, if the server provides one.
    if 'terms' in response_info and response_info['terms']:
        chip.logger.info('Terms of Service info for this server:')
        for line in response_info['terms'].splitlines():
            if line:
                chip.logger.info(line)


###################################
def remote_ping(chip):
    '''
    Helper method to call check_server on server
    '''

    # Make the request and print its response.
    response_info = _remote_ping(chip)

    # Print status value.
    server_status = response_info['status']
    chip.logger.info(f'Server status: {server_status}')
    if server_status != 'ready':
        chip.logger.warning('  Status is not "ready", server cannot accept new jobs.')

    # Print server-side version info.
    version_info = response_info['versions']
    version_suffix = ' version'
    max_name_string_len = max([len(s) for s in version_info.keys()]) + len(version_suffix)
    chip.logger.info('Server software versions:')
    for name, version in version_info.items():
        print_name = f'{name}{version_suffix}'
        chip.logger.info(f'  {print_name: <{max_name_string_len}}: {version}')

    # Print user info if applicable.
    if 'user_info' in response_info:
        user_info = response_info['user_info']
        if ('compute_time' not in user_info) or \
           ('bandwidth_kb' not in user_info):
            chip.logger.info('Error fetching user information from the remote server.')
            raise ValueError(f'Server response is not valid or missing fields: {user_info}')

        remote_cfg = get_remote_config(chip, False)
        if 'username' in remote_cfg:
            # Print the user's account info, and return.
            chip.logger.info(f'User {remote_cfg["username"]}:')

        time_remaining = user_info["compute_time"] / 60.0
        bandwidth_remaining = user_info["bandwidth_kb"]
        chip.logger.info(f'  Remaining compute time: {(time_remaining):.2f} minutes')
        chip.logger.info(f'  Remaining results bandwidth: {bandwidth_remaining} KiB')

    __print_tos(chip, response_info)

    # Return the response info in case the caller wants to inspect it.
    return response_info


def configure_server(chip, server=None, port=None, username=None, password=None):

    def confirm_dialog(message):
        confirmed = False
        while not confirmed:
            oin = input(f'{message} y/N: ')
            if (not oin) or (oin == 'n') or (oin == 'N'):
                return False
            elif (oin == 'y') or (oin == 'Y'):
                return True
        return False

    default_server_name = urllib.parse.urlparse(default_server).hostname

    # Find the config file/directory path.
    cfg_file = get_remote_config_file(chip, False)
    cfg_dir = os.path.dirname(cfg_file)

    # Create directory if it doesn't exist.
    if cfg_dir and not os.path.isdir(cfg_dir):
        os.makedirs(cfg_dir, exist_ok=True)

    # If an existing config file exists, prompt the user to overwrite it.
    if os.path.isfile(cfg_file):
        if not confirm_dialog('Overwrite existing remote configuration?'):
            return

    config = {}

    # If a command-line argument is passed in, use that as a public server address.
    if server:
        srv_addr = server
        chip.logger.info(f'Creating remote configuration file for server: {srv_addr}')
    else:
        # If no arguments were passed in, interactively request credentials from the user.
        srv_addr = input('Remote server address (leave blank to use default server):\n')
        srv_addr = srv_addr.replace(" ", "")

    if not srv_addr:
        srv_addr = default_server
        chip.logger.info(f'Using {srv_addr} as server')

    server = urllib.parse.urlparse(srv_addr)
    has_scheme = True
    if not server.hostname:
        # fake add a scheme to the url
        has_scheme = False
        server = urllib.parse.urlparse('https://' + srv_addr)
    if not server.hostname:
        raise ValueError(f'Invalid address provided: {srv_addr}')

    if has_scheme:
        config['address'] = f'{server.scheme}://{server.hostname}'
    else:
        config['address'] = server.hostname

    public_server = default_server_name in srv_addr
    if public_server and not confirm_dialog(__tos_str):
        return

    if server.port is not None:
        config['port'] = server.port

    if not public_server:
        if username is None:
            username = server.username
            if username is None:
                username = input('Remote username (leave blank for no username):\n')
                username = username.replace(" ", "")
        if password is None:
            password = server.password
            if password is None:
                password = input('Remote password (leave blank for no password):\n')
                password = password.replace(" ", "")

        if username:
            config['username'] = username
        if password:
            config['password'] = password

    config['directory_whitelist'] = []

    # Save the values to the target config file in JSON format.
    with open(cfg_file, 'w') as f:
        f.write(json.dumps(config, indent=4))

    # Let the user know that we finished successfully.
    chip.logger.info(f'Remote configuration saved to: {cfg_file}')


def configure_whitelist(chip, add, remove):
    try:
        cfg_file = get_remote_config_file(chip)
    except SiliconCompilerError as e:
        chip.logger.error(f'{e}')

    chip.logger.info(f'Updating credentials: {cfg_file}')
    cfg = get_remote_config(chip, True)

    if 'directory_whitelist' not in cfg:
        cfg['directory_whitelist'] = []

    if add:
        for path in add:
            path = os.path.abspath(path)
            chip.logger.info(f'Adding {path}')
            cfg['directory_whitelist'].append(path)

    if remove:
        for path in remove:
            path = os.path.abspath(path)
            if path in cfg['directory_whitelist']:
                chip.logger.info(f'Removing {path}')
                cfg['directory_whitelist'].remove(path)

    cfg['directory_whitelist'] = list(set(cfg['directory_whitelist']))

    # Save the values to the target config file in JSON format.
    with open(cfg_file, 'w') as f:
        f.write(json.dumps(cfg, indent=4))


def configure_print(chip):
    cfg = get_remote_config(chip, True)

    chip.logger.info(f'Server: {get_base_url(chip)}')
    if 'username' in cfg:
        chip.logger.info(f'Username: {cfg["username"]}')
    if 'directory_whitelist' in cfg and cfg['directory_whitelist']:
        chip.logger.info('Directory whitelist:')
        for path in sorted(cfg['directory_whitelist']):
            chip.logger.info(f'  {path}')



import os
import re
from siliconcompiler import sc_open, SiliconCompilerError
from siliconcompiler.tools.opensta import setup as tool_setup
from siliconcompiler.tools.opensta import runtime_options as tool_runtime_options
from siliconcompiler.tools._common import input_provides, add_common_file, \
    get_tool_task, record_metric
from siliconcompiler.tools._common.asic import set_tool_task_var, get_timing_modes


def setup(chip):
    '''
    Generate a static timing reports.
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    tool_setup(chip)

    chip.set('tool', tool, 'task', task, 'script', 'sc_timing.tcl',
             step=step, index=index, clobber=False)

    chip.set('tool', tool, 'task', task, 'threads', os.cpu_count(),
             step=step, index=index)

    design = chip.top()
    if f'{design}.vg' in input_provides(chip, step, index):
        chip.set('tool', tool, 'task', task, 'input', f'{design}.vg',
                 step=step, index=index)
    else:
        chip.set('tool', tool, 'task', task, 'require', 'input,netlist,verilog',
                 step=step, index=index)

    if f'{design}.sdc' in input_provides(chip, step, index):
        chip.set('tool', tool, 'task', task, 'input', f'{design}.sdc',
                 step=step, index=index)
    elif chip.valid('input', 'constraint', 'sdc') and \
            chip.get('input', 'constraint', 'sdc', step=step, index=index):
        chip.add('tool', tool, 'task', task, 'require', 'input,constraint,sdc',
                 step=step, index=index)

    set_tool_task_var(chip, param_key='top_n_paths',
                      default_value='10',
                      schelp='number of paths to report timing for')

    modes = get_timing_modes(chip)

    set_tool_task_var(chip, param_key='timing_mode',
                      default_value=modes[0],
                      schelp='timing mode to use')

    timing_mode = chip.get('tool', tool, 'task', task, 'var', 'timing_mode',
                           step=step, index=index)[0]
    if timing_mode not in modes:
        raise SiliconCompilerError(
            f'{timing_mode} mode is not present in timing constraints', chip=chip)

    for scenario in chip.getkeys('constraint', 'timing'):
        if chip.get('constraint', 'timing', scenario, 'mode',
                    step=step, index=index) != timing_mode:
            continue
        if chip.get('constraint', 'timing', scenario, 'file', step=step, index=index):
            chip.add('tool', tool, 'task', task, 'require', f'constraint,timing,{scenario},file',
                     step=step, index=index)

    add_common_file(chip, 'opensta_generic_sdc', 'sdc/sc_constraints.sdc')


def __report_map(chip, metric, basefile):
    corners = chip.getkeys('constraint', 'timing')
    mapping = {
        "power": [f"reports/power.{corner}.rpt" for corner in corners],
        "unconstrained": ["reports/unconstrained.rpt", "reports/unconstrained.topN.rpt"],
        "setuppaths": ["reports/setup.rpt", "reports/setup.topN.rpt"],
        "holdpaths": ["reports/hold.rpt", "reports/hold.topN.rpt"],
        "holdslack": ["reports/hold.rpt", "reports/hold.topN.rpt"],
        "setupslack": ["reports/setup.rpt", "reports/setup.topN.rpt"],
        "setuptns": ["reports/setup.rpt", "reports/setup.topN.rpt"],
        "holdtns": ["reports/hold.rpt", "reports/hold.topN.rpt"],
        "setupskew": ["reports/skew.setup.rpt", "reports/setup.rpt", "reports/setup.topN.rpt"],
        "holdskew": ["reports/skew.hold.rpt", "reports/hold.rpt", "reports/hold.topN.rpt"]
    }

    if metric in mapping:
        paths = [basefile]
        for path in mapping[metric]:
            if os.path.exists(path):
                paths.append(path)
        return paths
    return [basefile]


################################
# Post_process (post executable)
################################
def post_process(chip):
    '''
    Tool specific function to run after step execution
    '''

    # Check log file for errors and statistics
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    logfile = f"{step}.log"

    peakpower = []
    leakagepower = []
    skews = {}
    # parsing log file
    with sc_open(logfile) as f:
        timescale = "s"
        metric = None
        for line in f:
            metricmatch = re.search(r'^SC_METRIC:\s+(\w+)', line)
            value = re.search(r'(\d*\.?\d)*', line)
            fmax = re.search(r'fmax = (\d*\.?\d*)', line)
            tns = re.search(r'^tns (.*)', line)
            slack = re.search(r'^worst slack (.*)', line)
            skew = re.search(r'^\s*(.*)\s+(.*) skew', line)
            power = re.search(r'^Total(.*)', line)
            if metricmatch:
                metric = metricmatch.group(1)
                continue

            if metric:
                if metric == 'timeunit':
                    timescale = f'{line.strip()}s'
                    metric = None
                if metric == 'fmax':
                    if fmax:
                        record_metric(chip, step, index, 'fmax', float(fmax.group(1)),
                                      __report_map(chip, 'fmax', logfile),
                                      source_unit='MHz')
                        metric = None
                elif metric == 'power':
                    if power:
                        powerlist = power.group(1).split()
                        leakage = powerlist[2]
                        total = powerlist[3]

                        peakpower.append(float(total))
                        leakagepower.append(float(leakage))

                        metric = None
                elif metric == 'cellarea':
                    record_metric(chip, step, index, 'cellarea', float(value.group(0)),
                                  __report_map(chip, 'cellarea', logfile),
                                  source_unit='um^2')
                    metric = None
                elif metric in ('logicdepth',
                                'cells',
                                'nets',
                                'buffers',
                                'inverters',
                                'registers',
                                'unconstrained',
                                'pins',
                                'setuppaths',
                                'holdpaths'):
                    record_metric(chip, step, index, metric, int(value.group(0)),
                                  __report_map(chip, metric, logfile))
                    metric = None
                elif metric in ('holdslack', 'setupslack'):
                    if slack:
                        record_metric(chip, step, index, metric, float(slack.group(1)),
                                      __report_map(chip, metric, logfile),
                                      source_unit=timescale)
                        metric = None
                elif metric in ('setuptns', 'holdtns'):
                    if tns:
                        record_metric(chip, step, index, metric, float(tns.group(1)),
                                      __report_map(chip, metric, logfile),
                                      source_unit=timescale)
                        metric = None
                elif metric in ('setupskew', 'holdskew'):
                    if skew:
                        skews.setdefault(skew.group(2), []).append(float(skew.group(1)))
                else:
                    metric = None

    if peakpower:
        record_metric(chip, step, index, 'peakpower', max(peakpower),
                      __report_map(chip, 'peakpower', logfile),
                      source_unit='W')
    if leakagepower:
        record_metric(chip, step, index, 'leakagepower', max(leakagepower),
                      __report_map(chip, 'leakagepower', logfile),
                      source_unit='W')
    if skews:
        for skewtype, values in skews.items():
            skew = f'{skewtype}skew'
            record_metric(chip, step, index, skew, max(values),
                          __report_map(chip, skew, logfile),
                          source_unit=timescale)

    drv_report = "reports/drv_violators.rpt"
    if os.path.exists(drv_report):
        drv_count = 0
        with sc_open(drv_report) as f:
            for line in f:
                if re.search(r'\(VIOLATED\)$', line):
                    drv_count += 1

        record_metric(chip, step, index, 'drvs', drv_count,
                      [drv_report, logfile])


################################
# Runtime options
################################
def runtime_options(chip):
    return tool_runtime_options(chip)



import siliconcompiler
from siliconcompiler import SiliconCompilerError
from siliconcompiler.fpgas import lattice_ice40
from siliconcompiler.fpgas import vpr_example

from siliconcompiler.flows import fpgaflow


############################################################################
# DOCS
############################################################################
def make_docs(chip):
    chip.set('fpga', 'partname', 'ice40up5k-sg48')


####################################################
# Target Setup
####################################################
def setup(chip):
    '''
    Demonstration target for running the open-source fpgaflow.
    '''

    # 1. Configure fpga part
    part_name = chip.get('fpga', 'partname')
    if not part_name:
        raise SiliconCompilerError('FPGA partname has not been set.', chip=chip)

    # 2.  Load all available FPGAs
    chip.use(lattice_ice40)
    chip.use(vpr_example)

    if part_name not in chip.getkeys('fpga'):
        raise SiliconCompilerError(f'{part_name} has not been loaded', chip=chip)

    # 3. Load flow
    chip.use(fpgaflow, partname=part_name)

    # 4. Select default flow
    chip.set('option', 'flow', 'fpgaflow', clobber=False)


#########################
if __name__ == "__main__":
    target = siliconcompiler.Chip('<target>')
    setup(target)
    target.write_manifest('fpgaflow_demo.json')



from siliconcompiler.tools.builtin import _common
import os
from siliconcompiler import sc_open, SiliconCompilerError
from siliconcompiler import utils
from siliconcompiler.tools._common import input_provides, input_file_node_name, get_tool_task
from siliconcompiler import flowgraph
from siliconcompiler import scheduler


def make_docs(chip):
    from siliconcompiler.flows._common import _make_docs
    _make_docs(chip)
    chip.set('option', 'flow', 'asicflow')

    for step, index in flowgraph._get_flowgraph_entry_nodes(chip, 'asicflow'):
        scheduler._setup_node(chip, step, index)

    chip.set('arg', 'step', 'combine')
    chip.set('arg', 'index', '0')
    setup(chip)

    return chip


def setup(chip):
    '''
    A file concatenation pass that merges input files into a single set of outputs.
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    input_nodes = set()
    for nodes in input_provides(chip, step, index).values():
        input_nodes.update(nodes)

    if not input_nodes:
        raise SiliconCompilerError("Concatenate will not receive anything")
    if len(input_nodes) == 1:
        # nothing to concate to so remove
        return "no need to concatenate file"

    chip.set('tool', tool, 'task', task, 'input', [], step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', [], step=step, index=index)
    for file, nodes in input_provides(chip, step, index).items():
        chip.add('tool', tool, 'task', task, 'output', file,
                 step=step, index=index)
        for in_step, in_index in nodes:
            chip.add('tool', tool, 'task', task, 'input',
                     input_file_node_name(file, in_step, in_index),
                     step=step, index=index)


def _select_inputs(chip, step, index):
    return _common._select_inputs(chip, step, index)


def _gather_outputs(chip, step, index):
    '''Return set of filenames that are guaranteed to be in outputs
    directory after a successful run of step/index.'''

    flow = chip.get('option', 'flow')

    in_nodes = chip.get('flowgraph', flow, step, index, 'input')
    in_task_outputs = [flowgraph._gather_outputs(chip, *node) for node in in_nodes]

    if len(in_task_outputs) > 0:
        return in_task_outputs[0].union(*in_task_outputs[1:])

    return []


def run(chip):
    return _common.run(chip)


def post_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    def comment_flag(ext):
        if ext.startswith('tcl'):
            return '#'
        if ext.startswith('v'):
            return '//'
        return None

    for file, nodes in input_provides(chip, step, index).items():
        comment = comment_flag(utils.get_file_ext(file))
        with open(f'outputs/{file}', 'w') as out:
            for in_step, in_index in nodes:
                ifile = f'inputs/{input_file_node_name(file, in_step, in_index)}'
                if not os.path.isfile(ifile):
                    continue
                with sc_open(ifile) as in_file:
                    if comment:
                        out.write(f'{comment} Start of SiliconCompiler input: {ifile}\n')
                    out.write(in_file.read())
                    if comment:
                        out.write(f'{comment} End of SiliconCompiler input: {ifile}\n\n')



from siliconcompiler.tools.builtin import _common
from siliconcompiler.schema import Schema
from siliconcompiler.scheduler import _haltstep
from siliconcompiler.tools.builtin.builtin import set_io_files
from siliconcompiler import utils, flowgraph, SiliconCompilerError

import re


def setup(chip):
    '''
    Tests an assertion on an input task.

    The input to this task is verified to ensure that all assertions
    are True. If any of the assertions fail, False is returned.
    Assertions are passed in using ['flowgraph', flow, step, index, 'args'] in the form
    'metric==0.0'.
    The allowed conditional operators are: >, <, >=, <=, ==
    '''

    set_io_files(chip, outputs=False)


def _select_inputs(chip, step, index):
    inputs = _common._select_inputs(chip, step, index)
    if len(inputs) != 1:
        raise SiliconCompilerError(
            f'{step}{index} receives {len(inputs)} inputs, but only supports one', chip=chip)
    inputs = inputs[0]
    flow = chip.get('option', 'flow')
    arguments = chip.get('flowgraph', flow, step, index, 'args')

    if len(arguments) == 0:
        raise SiliconCompilerError(f'{step}{index} requires arguments for verify', chip=chip)

    passes = True
    for criteria in arguments:
        m = re.match(r'(\w+)([\>\=\<]+)(\w+)', criteria)
        if not m:
            raise SiliconCompilerError(f"Illegal verify criteria: {criteria}", chip=chip)

        metric = m.group(1)
        op = m.group(2)
        goal = m.group(3)
        if metric not in chip.getkeys('metric'):
            raise SiliconCompilerError(
                f"Criteria must use legal metrics only: {criteria}", chip=chip)

        value = chip.get('metric', metric, step=inputs[0], index=inputs[1])

        if value is None:
            raise SiliconCompilerError(
                f"Missing metric for {metric} in {inputs[0]}{inputs[1]}", chip=chip)

        metric_type = chip.get('metric', metric, field='type')
        goal = Schema._normalize_value(goal, metric_type, "", None)
        if not utils.safecompare(chip, value, op, goal):
            chip.error(f"{step}{index} fails '{metric}' metric: {value}{op}{goal}")

    if not passes:
        _haltstep(chip, flow, step, index)

    return inputs


def _gather_outputs(chip, step, index):
    flow = chip.get('option', 'flow')

    in_nodes = chip.get('flowgraph', flow, step, index, 'input')
    in_task_outputs = [flowgraph._gather_outputs(chip, *node) for node in in_nodes]

    if len(in_task_outputs) > 0:
        return in_task_outputs[0].intersection(*in_task_outputs[1:])

    return []


def run(chip):
    return _common.run(chip)


def post_process(chip):
    _common.post_process(chip)



# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.
import sys
import siliconcompiler
import os
from siliconcompiler.apps._common import pick_manifest, manifest_switches, UNSET_DESIGN


def main():
    progname = "sc-dashboard"
    description = """
-----------------------------------------------------------
SC app to open a dashboard for a given manifest.

To open and allow sc-dashboard to autoload manifest:
    sc-dashboard

To open by specifying manifest:
    sc-dashboard -cfg <path to manifest>

To open by specifying design and optionally jobname:
    sc-dashboard -design <name>
    sc-dashboard -design <name> -jobname <jobname>

To specify a different port than the default:
    sc-dashboard -cfg <path to manifest> -port 10000

To include another chip object to compare to:
    sc-dashboard -cfg <path to manifest> -graph_cfg <name of manifest> <path to other manifest>
        -graph_cfg <path to other manifest> ...
-----------------------------------------------------------
"""

    # Create a base chip class.
    chip = siliconcompiler.Chip(UNSET_DESIGN)

    dashboard_arguments = {
        "-port": {'type': int,
                  'help': 'port to open the dashboard app on',
                  'metavar': '<port>',
                  'sc_print': False},
        "-graph_cfg": {'type': str,
                       'nargs': '+',
                       'action': 'append',
                       'help': 'chip name - optional, path to chip manifest (json)',
                       'metavar': '<[manifest name, manifest path>',
                       'sc_print': False}
    }

    try:
        switches = chip.create_cmdline(
            progname,
            switchlist=[*manifest_switches(),
                        '-loglevel'],
            description=description,
            additional_args=dashboard_arguments)
    except Exception as e:
        chip.logger.error(e)
        return 1

    if not chip.get('option', 'cfg'):
        manifest = pick_manifest(chip)

        if manifest:
            chip.logger.info(f'Loading manifest: {manifest}')
            chip.read_manifest(manifest)

    # Error checking
    design = chip.get('design')
    if design == UNSET_DESIGN:
        chip.logger.error('Design not loaded')
        return 1

    graph_chips = []
    if switches['graph_cfg']:
        for i, name_and_file_path in enumerate(switches['graph_cfg']):
            args = len(name_and_file_path)
            if args == 0:
                continue
            elif args == 1:
                name = f'cfg{i}'
                file_path = name_and_file_path[0]
            elif args == 2:
                name = name_and_file_path[0]
                file_path = name_and_file_path[1]
            else:
                raise ValueError(('graph_cfg accepts a max of 2 values, you supplied'
                                  f' {args} in "-graph_cfg {name_and_file_path}"'))
            if not os.path.isfile(file_path):
                raise ValueError(f'not a valid file path : {file_path}')
            graph_chip = siliconcompiler.core.Chip(design='')
            graph_chip.read_manifest(file_path)
            graph_chips.append({
                'chip': graph_chip,
                'name': name,
                'cfg_path': os.path.abspath(file_path)
            })

    chip.dashboard(wait=True, port=switches['port'], graph_chips=graph_chips)

    return 0


#########################
if __name__ == "__main__":
    sys.exit(main())



import re

from siliconcompiler.tools.yosys.yosys import setup as setup_tool
from siliconcompiler.tools.yosys.syn_asic import setup_asic
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric


def setup(chip):
    '''
    Perform logical equivalence checks
    '''

    # Generic tool setup.
    setup_tool(chip)

    # Generic setup.
    setup_asic(chip)

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    design = chip.top()

    # Set yosys script path.
    chip.set('tool', tool, 'task', task, 'script', 'sc_lec.tcl',
             step=step, index=index, clobber=False)

    # Input/output requirements.
    if not chip.valid('input', 'netlist', 'verilog') or \
       not chip.get('input', 'netlist', 'verilog', step=step, index=index):
        chip.set('tool', tool, 'task', task, 'input', design + '.vg', step=step, index=index)
    # if not chip.get('input', 'rtl', 'verilog'):
        # TODO: Not sure this logic makes sense? Seems like reverse of tcl
        # chip.set('tool', tool, 'task', task, 'input', design + '.v', step=step, index=index)


##################################################
def post_process(chip):
    ''' Tool specific function to run after step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    with sc_open(step + ".log") as f:
        for line in f:
            if line.endswith('Equivalence successfully proven!\n'):
                record_metric(chip, step, index, 'drvs', 0, step + ".log")
                continue

            errors = re.search(r'Found a total of (\d+) unproven \$equiv cells.', line)
            if errors is not None:
                num_errors = int(errors.group(1))
                record_metric(chip, step, index, 'drvs', num_errors, step + ".log")



from siliconcompiler import Flow

from siliconcompiler.tools.klayout import drc


def setup():
    '''
    Perform a DRC run on an input GDS
    '''
    flow = Flow('drcflow')
    flow.node('drcflow', 'drc', drc)

    return flow




from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools.openroad.openroad import set_pnr_inputs, set_pnr_outputs


def setup(chip):
    '''
    Not implemented yet
    '''

    # Generic tool setup.
    setup_tool(chip)

    set_pnr_inputs(chip)
    set_pnr_outputs(chip)


def pre_process(chip):
    or_pre_process(chip)
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



from siliconcompiler.tools._common import input_provides, get_tool_task
from siliconcompiler.tools.xyce import setup as tool_setup
import os


def setup(chip):
    tool_setup(chip)

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)

    design = chip.top()

    if f'{design}.xyce' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.xyce',
                 step=step, index=index)
    elif f'{design}.cir' in input_provides(chip, step, index):
        chip.add('tool', tool, 'task', task, 'input', f'{design}.cir',
                 step=step, index=index)
    else:
        chip.add('tool', tool, 'task', task, 'require', 'input,netlist,spice',
                 step=step, index=index)

    chip.set('tool', tool, 'task', task, 'var', 'trace', 'false',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'trace', 'true/false, enable dumping all signals',
             field='help')

    allowed_traced = ['ASCII', 'binary']
    chip.set('tool', tool, 'task', task, 'var', 'trace_format', 'ASCII',
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'var', 'trace_format', 'Format to use for traces. '
             f'Allowed values are {allowed_traced}',
             field='help')

    if chip.get('tool', tool, 'task', task, 'var', 'trace_format', step=step, index=index)[0] \
            not in allowed_traced:
        raise ValueError(f"{chip.get('tool', tool, 'task', task, 'var', 'trace_format')[0]} "
                         "is not an accepted value")

    if chip.get('tool', tool, 'task', task, 'var', 'trace_format', step=step, index=index) == \
            ['ASCII']:
        chip.add('tool', tool, 'task', task, 'option', '-a',
                 step=step, index=index)

    if chip.get('tool', tool, 'task', task, 'var', 'trace',
                step=step, index=index) == ['true']:
        chip.add('tool', tool, 'task', task, 'output', f'{design}.raw',
                 step=step, index=index)
        chip.add('tool', tool, 'task', task, 'option', f'-r outputs/{design}.raw',
                 step=step, index=index)


def runtime_options(chip):
    design = chip.top()
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    if f'{design}.xyce' in input_provides(chip, step, index):
        if os.path.isfile(f'inputs/{design}.xyce'):
            return [f'inputs/{design}.xyce']
        elif os.path.isfile(f'inputs/{design}.xyce/{design}.cir'):
            return [f'inputs/{design}.xyce/{design}.cir']
        else:
            return [f'inputs/{design}.xyce']
    elif f'{design}.cir' in input_provides(chip, step, index):
        return [f'inputs/{design}.cir']

    return chip.find_files('input', 'netlist', 'spice', step=step, index=index)



'''
Yosys is a framework for RTL synthesis that takes synthesizable
Verilog-2005 design and converts it to BLIF, EDIF, BTOR, SMT,
Verilog netlist etc. The tool supports logical synthesis and
tech mapping to ASIC standard cell libraries, FPGA architectures.
In addition it has built in formal methods for property and
equivalence checking.

Documentation: https://yosyshq.readthedocs.io/projects/yosys/en/latest/

Sources: https://github.com/YosysHQ/yosys

Installation: https://github.com/YosysHQ/yosys
'''

import re
import json
from siliconcompiler import sc_open
from siliconcompiler.tools._common import get_tool_task, record_metric
from siliconcompiler.targets import asap7_demo


######################################################################
# Make Docs
######################################################################
def make_docs(chip):
    chip.use(asap7_demo)


################################
# Setup Tool (pre executable)
################################
def setup(chip):
    ''' Tool specific function to run before step execution
    '''

    # If the 'lock' bit is set, don't reconfigure.
    tool = 'yosys'
    refdir = 'tools/' + tool
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    # Standard Setup
    chip.set('tool', tool, 'exe', 'yosys')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=0.41', clobber=False)
    chip.set('tool', tool, 'format', 'tcl', clobber=False)

    # Task Setup
    # common to all
    option = []
    if chip.get('option', 'breakpoint', step=step, index=index):
        option.append('-C')
    option.append('-c')
    chip.set('tool', tool, 'task', task, 'option', option, step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'refdir', refdir, step=step, index=index,
             package='siliconcompiler', clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'warnings', "Warning:",
             step=step, index=index, clobber=False)
    chip.set('tool', tool, 'task', task, 'regex', 'errors', "^ERROR",
             step=step, index=index, clobber=False)


################################
# Version Check
################################
def parse_version(stdout):
    # Yosys 0.9+3672 (git sha1 014c7e26, gcc 7.5.0-3ubuntu1~18.04 -fPIC -Os)
    return stdout.split()[1]


def normalize_version(version):
    # Replace '+', which represents a "local version label", with '-', which is
    # an "implicit post release number".
    return version.replace('+', '-')


def syn_setup(chip):
    ''' Helper method for configs specific to synthesis tasks.
    '''

    # Generic tool setup.
    setup(chip)

    tool = 'yosys'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)
    design = chip.top()

    # Set yosys script path.
    chip.set('tool', tool, 'task', task, 'script', 'sc_syn.tcl',
             step=step, index=index, clobber=False)

    # Input/output requirements.
    chip.set('tool', tool, 'task', task, 'input', design + '.v', step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', design + '.vg', step=step, index=index)


##################################################
def syn_post_process(chip):
    ''' Tool specific function to run after step execution
    '''

    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')

    with sc_open("reports/stat.json") as f:
        metrics = json.load(f)
        if "design" in metrics:
            metrics = metrics["design"]

        if "area" in metrics:
            record_metric(chip, step, index, 'cellarea',
                          float(metrics["area"]),
                          "reports/stat.json",
                          source_unit='um^2')
        if "num_cells" in metrics:
            record_metric(chip, step, index, 'cells',
                          metrics["num_cells"],
                          "reports/stat.json")
        if "num_wire_bits" in metrics:
            record_metric(chip, step, index, 'nets',
                          metrics["num_wire_bits"],
                          "reports/stat.json")
        if "num_port_bits" in metrics:
            record_metric(chip, step, index, 'pins',
                          metrics["num_port_bits"],
                          "reports/stat.json")

    registers = None
    with sc_open(f"{step}.log") as f:
        for line in f:
            line_registers = re.findall(r"^\s*mapped ([0-9]+) \$_DFF.*", line)
            if line_registers:
                if registers is None:
                    registers = 0
                registers += int(line_registers[0])
    if registers is not None:
        record_metric(chip, step, index, 'registers', registers, f"{step}.log")


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("yosys.json")



import siliconcompiler

from siliconcompiler.tools.magic import extspice
from siliconcompiler.tools.magic import drc
from siliconcompiler.tools.netgen import lvs
from siliconcompiler.tools.builtin import nop
from siliconcompiler.tools.builtin import join


def make_docs(chip):
    chip.set('input', 'netlist', 'verilog', 'test')
    return setup()


def setup():
    '''A flow for running LVS/DRC signoff on a GDS layout.

    Inputs must be passed to this flow as follows::

        flow.input('<path-to-layout>.gds')
        flow.input('<path-to-netlist>.vg')
    '''
    flowname = 'signoffflow'

    flow = siliconcompiler.Flow(flowname)

    # nop import since we don't need to pull in any sources
    flow.node(flowname, 'import', nop)

    flow.node(flowname, 'extspice', extspice)
    flow.node(flowname, 'drc', drc)
    flow.node(flowname, 'lvs', lvs)
    flow.node(flowname, 'signoff', join)

    flow.edge(flowname, 'import', 'drc')
    flow.edge(flowname, 'import', 'extspice')
    flow.edge(flowname, 'extspice', 'lvs')
    flow.edge(flowname, 'lvs', 'signoff')
    flow.edge(flowname, 'drc', 'signoff')

    # Set default goal
    for step in flow.getkeys('flowgraph', flowname):
        flow.set('flowgraph', flowname, step, '0', 'goal', 'errors', 0)

    return flow


##################################################
if __name__ == "__main__":
    chip = siliconcompiler.Chip('design')
    flow = make_docs(chip)
    chip.use(flow)
    chip.write_flowgraph(f"{flow.top()}.png", flow=flow.top())




from siliconcompiler import NodeStatus, SiliconCompilerError
from siliconcompiler import utils
import shutil
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.flowgraph import _get_pruned_node_inputs


###########################################################################
def _mux(chip, *nodes, operations=None):
    '''
    Shared function used for min and max calculation.
    '''

    nodelist = list(nodes)

    # Keeping track of the steps/indexes that have goals met
    failed = {}
    for step, index in nodelist:
        if step not in failed:
            failed[step] = {}
        failed[step][index] = False

        if chip.get('record', 'status', step=step, index=index) == NodeStatus.ERROR:
            failed[step][index] = True
        else:
            failed[step][index] = False

    candidates = [(step, index) for step, index in nodelist if not failed[step][index]]
    best_score = 0
    for metric, op in operations:
        if op not in ('minimum', 'maximum'):
            raise ValueError('Invalid op')

        values = [chip.get('metric', metric, step=step, index=index) for step, index in candidates]

        if op == 'minimum':
            target = min(values)
        else:
            target = max(values)

        winners = []
        for value, node in zip(values, candidates):
            if value == target:
                winners.append(node)
        candidates = winners

        if len(candidates) == 1:
            break

    if len(candidates) == 0:
        # Restore step list and pick first step
        candidates = nodelist

    return (best_score, candidates[0])


###########################################################################
def _minmax(chip, *nodes, op=None):
    '''
    Shared function used for min and max calculation.
    '''

    if op not in ('minimum', 'maximum'):
        raise ValueError('Invalid op')

    flow = chip.get('option', 'flow')
    nodelist = list(nodes)

    # Keeping track of the steps/indexes that have goals met
    failed = {}
    for step, index in nodelist:
        if step not in failed:
            failed[step] = {}
        failed[step][index] = False

        if chip.get('record', 'status', step=step, index=index) == NodeStatus.ERROR:
            failed[step][index] = True
        else:
            for metric in chip.getkeys('metric'):
                if chip.valid('flowgraph', flow, step, index, 'goal', metric):
                    goal = chip.get('flowgraph', flow, step, index, 'goal', metric)
                    real = chip.get('metric', metric, step=step, index=index)
                    if real is None:
                        raise SiliconCompilerError(
                            f'Metric {metric} has goal for {step}{index} '
                            'but it has not been set.', chip=chip)
                    if abs(real) > goal:
                        chip.logger.warning(f"Step {step}{index} failed "
                                            f"because it didn't meet goals for '{metric}' "
                                            "metric.")
                        failed[step][index] = True

    # Calculate max/min values for each metric
    max_val = {}
    min_val = {}
    for metric in chip.getkeys('metric'):
        max_val[metric] = 0
        min_val[metric] = float("inf")
        for step, index in nodelist:
            if not failed[step][index]:
                real = chip.get('metric', metric, step=step, index=index)
                if real is None:
                    continue
                max_val[metric] = max(max_val[metric], real)
                min_val[metric] = min(min_val[metric], real)

    # Select the minimum index
    best_score = float('inf') if op == 'minimum' else float('-inf')
    winner = None
    for step, index in nodelist:
        if failed[step][index]:
            continue

        score = 0.0
        for metric in chip.getkeys('flowgraph', flow, step, index, 'weight'):
            weight = chip.get('flowgraph', flow, step, index, 'weight', metric)
            if not weight:
                # skip if weight is 0 or None
                continue

            real = chip.get('metric', metric, step=step, index=index)
            if real is None:
                raise SiliconCompilerError(
                    f'Metric {metric} has weight for {step}{index} '
                    'but it has not been set.', chip=chip)

            if not (max_val[metric] - min_val[metric]) == 0:
                scaled = (real - min_val[metric]) / (max_val[metric] - min_val[metric])
            else:
                scaled = max_val[metric]
            score = score + scaled * weight

        if ((op == 'minimum' and score < best_score) or (op == 'maximum' and score > best_score)):
            best_score = score
            winner = (step, index)

    return (best_score, winner)


def run(chip):
    return 0


def post_process(chip):
    shutil.copytree('inputs', 'outputs', dirs_exist_ok=True, copy_function=utils.link_symlink_copy)


def _select_inputs(chip, step, index):
    _, task = get_tool_task(chip, step, index)

    chip.logger.info(f"Running builtin task '{task}'")

    flow = chip.get('option', 'flow')
    return _get_pruned_node_inputs(chip, flow, (step, index))




from siliconcompiler.tools.openroad.openroad import setup as setup_tool
from siliconcompiler.tools.openroad.openroad import build_pex_corners
from siliconcompiler.tools.openroad.openroad import post_process as or_post_process
from siliconcompiler.tools.openroad.openroad import pre_process as or_pre_process
from siliconcompiler.tools.openroad.openroad import _set_reports, set_pnr_inputs, set_pnr_outputs
from siliconcompiler.tools._common import get_tool_task
from siliconcompiler.tools._common.asic import get_mainlib
from siliconcompiler import NodeStatus


def setup(chip):
    '''
    Design for manufacturing step will insert fill if specified
    '''

    # Generic tool setup.
    setup_tool(chip)

    set_pnr_inputs(chip)
    set_pnr_outputs(chip)

    _set_reports(chip, [
        'setup',
        'hold',
        'unconstrained',
        'clock_skew',
        'power',
        'drv_violations',
        'fmax',

        # Images
        'placement_density',
        'routing_congestion',
        'power_density',
        'clock_placement',
        'clock_trees',
        'optimization_placement'
    ])


def pre_process(chip):
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    tool, task = get_tool_task(chip, step, index)
    if not chip.find_files('tool', tool, 'task', task, 'prescript',
                           step=step, index=index) and \
            not chip.find_files('tool', tool, 'task', task, 'postscript',
                                step=step, index=index) and \
            chip.get('tool', tool, 'task', task, 'var', 'fin_add_fill',
                     step=step, index=index) == ["true"]:
        pdk = chip.get('option', 'pdk')
        stackup = chip.get('option', 'stackup')
        mainlib = get_mainlib(chip)
        libtype = chip.get('library', mainlib, 'asic', 'libarch', step=step, index=index)

        if not chip.find_files('pdk', pdk, 'aprtech', tool, stackup, libtype, 'fill'):
            chip.set('record', 'status', NodeStatus.SKIPPED, step=step, index=index)
            chip.logger.warning(f'{step}{index} will be skipped since there is nothing to do.')

    or_pre_process(chip)
    build_pex_corners(chip)


def post_process(chip):
    or_post_process(chip)



import contextlib
import hashlib
import os
import re
import psutil
import shutil
from pathlib import Path, PurePosixPath
from siliconcompiler._metadata import version as sc_version
from jinja2 import Environment, FileSystemLoader

import sys
if sys.version_info < (3, 10):
    from importlib_metadata import entry_points
else:
    from importlib.metadata import entry_points


PACKAGE_ROOT = os.path.dirname(os.path.dirname(os.path.abspath(__file__)))

_siliconcompiler_data_path = 'git+https://github.com/siliconcompiler/siliconcompiler'


def link_symlink_copy(srcfile, dstfile):
    # first try hard linking, then symbolic linking,
    # and finally just copy the file
    for method in [os.link, os.symlink, shutil.copy2]:
        try:
            # create link
            return method(srcfile, dstfile)
            # success, no need to continue trying
        except OSError:
            pass


def link_copy(srcfile, dstfile):
    # first try hard linking, then symbolic linking,
    # and finally just copy the file
    for method in [os.link, shutil.copy2]:
        try:
            # create link
            return method(srcfile, dstfile)
            # success, no need to continue trying
        except OSError:
            pass


def terminate_process(pid, timeout=3):
    '''Terminates a process and all its (grand+)children.

    Based on https://psutil.readthedocs.io/en/latest/#psutil.wait_procs and
    https://psutil.readthedocs.io/en/latest/#kill-process-tree.
    '''
    assert pid != os.getpid(), "won't terminate myself"
    parent = psutil.Process(pid)
    children = parent.children(recursive=True)
    children.append(parent)
    for p in children:
        try:
            p.terminate()
        except psutil.NoSuchProcess:
            # Process may have terminated on its own in the meantime
            pass

    _, alive = psutil.wait_procs(children, timeout=timeout)
    for p in alive:
        # If processes are still alive after timeout seconds, send more
        # aggressive signal.
        p.kill()


def get_file_ext(filename):
    '''Get base file extension for a given path, disregarding .gz.'''
    if filename.lower().endswith('.gz'):
        filename = os.path.splitext(filename)[0]
    filetype = os.path.splitext(filename)[1].lower().lstrip('.')
    return filetype


def get_default_iomap():
    """
    Default input file map for SC with filesets and extensions
    """

    # Record extensions:

    # High level languages
    hll_c = ('c', 'cc', 'cpp', 'c++', 'cp', 'cxx', 'hpp')
    hll_bsv = ('bsv',)
    hll_scala = ('scala',)
    hll_python = ('py',)

    config_chisel = ('sbt',)

    # Register transfer languages
    rtl_verilog = ('v', 'verilog')
    rtl_systemverilog = ('sv',)
    rtl_vhdl = ('vhd', 'vhdl')

    # Timing libraries
    timing_liberty = ('lib', 'ccs')

    # Layout
    layout_def = ('def',)
    layout_lef = ('lef',)
    layout_gds = ('gds', 'gds2', 'gdsii')
    layout_oas = ('oas', 'oasis')
    layout_gerber = ('gbr', 'gerber')
    layout_odb = ('odb',)

    # Netlist
    netlist_cdl = ('cdl',)
    netlist_sp = ('sp', 'spice')
    netlist_verilog = ('vg',)

    # Waveform
    waveform_vcd = ('vcd',)

    # Constraint
    constraint_sdc = ('sdc', )
    constraint_upf = ('upf', )

    # FPGA constraints
    fpga_xdc = ('xdc',)
    constraint_pcf = ('pcf',)
    fpga_vpr_place = ('place',)
    fpga_vpr_route = ('route',)

    # Build default map with fileset and type
    default_iomap = {}
    default_iomap.update({ext: ('hll', 'c') for ext in hll_c})
    default_iomap.update({ext: ('hll', 'bsv') for ext in hll_bsv})
    default_iomap.update({ext: ('hll', 'scala') for ext in hll_scala})
    default_iomap.update({ext: ('hll', 'python') for ext in hll_python})
    default_iomap.update({ext: ('config', 'chisel') for ext in config_chisel})

    default_iomap.update({ext: ('rtl', 'verilog') for ext in rtl_verilog})
    default_iomap.update({ext: ('rtl', 'systemverilog') for ext in rtl_systemverilog})
    default_iomap.update({ext: ('rtl', 'vhdl') for ext in rtl_vhdl})

    default_iomap.update({ext: ('timing', 'liberty') for ext in timing_liberty})

    default_iomap.update({ext: ('layout', 'def') for ext in layout_def})
    default_iomap.update({ext: ('layout', 'lef') for ext in layout_lef})
    default_iomap.update({ext: ('layout', 'gds') for ext in layout_gds})
    default_iomap.update({ext: ('layout', 'oas') for ext in layout_oas})
    default_iomap.update({ext: ('layout', 'gerber') for ext in layout_gerber})
    default_iomap.update({ext: ('layout', 'odb') for ext in layout_odb})

    default_iomap.update({ext: ('netlist', 'cdl') for ext in netlist_cdl})
    default_iomap.update({ext: ('netlist', 'sp') for ext in netlist_sp})
    default_iomap.update({ext: ('netlist', 'verilog') for ext in netlist_verilog})

    default_iomap.update({ext: ('waveform', 'vcd') for ext in waveform_vcd})

    default_iomap.update({ext: ('constraint', 'sdc') for ext in constraint_sdc})
    default_iomap.update({ext: ('constraint', 'upf') for ext in constraint_upf})
    default_iomap.update({ext: ('constraint', 'pcf') for ext in constraint_pcf})

    default_iomap.update({ext: ('fpga', 'xdc') for ext in fpga_xdc})
    default_iomap.update({ext: ('fpga', 'vpr_place') for ext in fpga_vpr_place})
    default_iomap.update({ext: ('fpga', 'vpr_route') for ext in fpga_vpr_route})

    return default_iomap


def format_fileset_type_table(indent=12):
    '''
    Generate a table to use in the __doc__ of the input function which auto
    updates based on the iomap
    '''
    table = "filetype      | fileset    | suffix (case insensitive)\n"
    indent = " " * indent
    table += f"{indent}--------------|------------|---------------------------------------------\n"

    iobytype = {}
    for ext, settype in get_default_iomap().items():
        fileset, filetype = settype
        iobytype.setdefault((fileset, filetype), []).append(ext)

    for settype, exts in iobytype.items():
        fileset, filetype = settype
        ext = ",".join(exts)
        table += f"{indent}{filetype:<14}| {fileset:<11}| {ext}\n"

    return table


def default_credentials_file():
    cfg_file = os.path.join(Path.home(), '.sc', 'credentials')

    return cfg_file


def default_cache_dir():
    cfg_file = os.path.join(Path.home(), '.sc', 'cache')

    return cfg_file


def default_email_credentials_file():
    cfg_file = os.path.join(Path.home(), '.sc', 'email.json')

    return cfg_file


def register_sc_data_source(chip):
    chip.register_source('siliconcompiler_data',
                         _siliconcompiler_data_path,
                         'v'+sc_version)


@contextlib.contextmanager
def sc_open(path, *args, **kwargs):
    kwargs['errors'] = 'ignore_with_warning'
    fobj = open(path, *args, **kwargs)
    try:
        with contextlib.closing(fobj):
            yield fobj
    finally:
        pass


def get_file_template(path, root=os.path.join(PACKAGE_ROOT, 'templates')):
    if os.path.isabs(path):
        root = os.path.dirname(path)
        path = os.path.basename(path)

    env = Environment(loader=FileSystemLoader(root))
    return env.get_template(path)


#######################################
def safecompare(chip, value, op, goal):
    # supported relational operations
    # >, >=, <=, <. ==, !=
    if op == ">":
        return bool(value > goal)
    elif op == ">=":
        return bool(value >= goal)
    elif op == "<":
        return bool(value < goal)
    elif op == "<=":
        return bool(value <= goal)
    elif op == "==":
        return bool(value == goal)
    elif op == "!=":
        return bool(value != goal)
    else:
        chip.error(f"Illegal comparison operation {op}")


###########################################################################
def grep(chip, args, line):
    """
    Emulates the Unix grep command on a string.

    Emulates the behavior of the Unix grep command that is etched into
    our muscle memory. Partially implemented, not all features supported.
    The function returns None if no match is found.

    Args:
        arg (string): Command line arguments for grep command
        line (string): Line to process

    Returns:
        Result of grep command (string).

    """

    # Quick return if input is None
    if line is None:
        return None

    # Partial list of supported grep options
    options = {
        '-v': False,  # Invert the sense of matching
        '-i': False,  # Ignore case distinctions in patterns and data
        '-E': False,  # Interpret PATTERNS as extended regular expressions.
        '-e': False,  # Safe interpretation of pattern starting with "-"
        '-x': False,  # Select only matches that exactly match the whole line.
        '-o': False,  # Print only the match parts of a matching line
        '-w': False}  # Select only lines containing matches that form whole words.

    # Split into repeating switches and everything else
    match = re.match(r'\s*((?:\-\w\s)*)(.*)', args)

    pattern = match.group(2)

    # Split space separated switch string into list
    switches = match.group(1).strip().split(' ')

    # Find special -e switch update the pattern
    for i in range(len(switches)):
        if switches[i] == "-e":
            if i != (len(switches)):
                pattern = ' '.join(switches[i + 1:]) + " " + pattern
                switches = switches[0:i + 1]
                break
            options["-e"] = True
        elif switches[i] in options.keys():
            options[switches[i]] = True
        elif switches[i] != '':
            chip.logger.error(switches[i])

    # REGEX
    # TODO: add all the other optinos
    match = re.search(rf"({pattern})", line)
    if bool(match) == bool(options["-v"]):
        return None
    else:
        return line


#######################################
def _resolve_env_vars(chip, filepath):
    if not filepath:
        return None

    env_save = os.environ.copy()
    for env in chip.getkeys('option', 'env'):
        os.environ[env] = chip.get('option', 'env', env)
    resolved_path = os.path.expandvars(filepath)
    os.environ.clear()
    os.environ.update(env_save)

    # variables that don't exist in environment get ignored by `expandvars`,
    # but we can do our own error checking to ensure this doesn't result in
    # silent bugs
    envvars = re.findall(r'\$(\w+)', resolved_path)
    for var in envvars:
        chip.logger.warning(f'Variable {var} in {filepath} not defined in environment')

    return resolved_path


###########################################################################
def find_sc_file(chip, filename, missing_ok=False, search_paths=None):
    """
    Returns the absolute path for the filename provided.

    Searches the for the filename provided and returns the absolute path.
    If no valid absolute path is found during the search, None is returned.

    Shell variables ('$' followed by strings consisting of numbers,
    underscores, and digits) are replaced with the variable value.

    Args:
        filename (str): Relative or absolute filename.
        missing_ok (bool): If False, error out if no valid absolute path
            found, rather than returning None.
        search_paths (list): List of directories to search under instead of
            the defaults.

    Returns:
        Returns absolute path of 'filename' if found, otherwise returns
        None.

    Examples:
        >>> chip._find_sc_file('flows/asicflow.py')
        Returns the absolute path based on the sc installation directory.

    """

    if not filename:
        return None

    # Replacing environment variables
    filename = _resolve_env_vars(chip, filename)

    # If we have an absolute path, pass-through here
    if os.path.isabs(filename) and os.path.exists(filename):
        return filename

    # Otherwise, search relative to search_paths
    if search_paths is None:
        search_paths = [chip.cwd]

    searchdirs = ', '.join([str(p) for p in search_paths])
    chip.logger.debug(f"Searching for file {filename} in {searchdirs}")

    result = None
    for searchdir in search_paths:
        if not os.path.isabs(searchdir):
            searchdir = os.path.join(chip.cwd, searchdir)

        abspath = os.path.abspath(os.path.join(searchdir, filename))
        if os.path.exists(abspath):
            result = abspath
            break

    if result is None and not missing_ok:
        chip.error(f"File {filename} was not found")

    return result


def get_plugins(system):
    '''
    Search for python modules with a specific function
    '''

    plugins = []
    discovered_plugins = entry_points(group=f'siliconcompiler.{system}')
    for plugin in discovered_plugins:
        plugins.append(plugin.load())

    return plugins


def truncate_text(text, width):
    if len(text) <= width:
        return text

    keep_end = 0
    while text[-1 - keep_end].isnumeric():
        if keep_end >= 2:
            break
        keep_end += 1

    while len(text) > width:
        break_at = len(text) - keep_end - 3
        text = text[:break_at-1] + '...' + text[break_at+3:]

    return text


def get_hashed_filename(path, package=None, hash=hashlib.sha1):
    '''
    Utility to map collected file to an unambiguous name based on its path.

    The mapping looks like:
    path/to/file.ext => file_<hash('path/to/file')>.ext
    '''
    path = PurePosixPath(path)
    ext = ''.join(path.suffixes)

    # strip off all file suffixes to get just the bare name
    barepath = path
    while barepath.suffix:
        barepath = PurePosixPath(barepath.stem)
    filename = str(barepath.parts[-1])

    if not package:
        package = ''
    else:
        package = f'{package}:'

    path_to_hash = f'{package}{str(path)}'

    pathhash = hash(path_to_hash.encode('utf-8')).hexdigest()

    return f'{filename}_{pathhash}{ext}'



from siliconcompiler.tools._common import get_tool_task


def setup(chip):
    '''
    Perform automated place and route on FPGAs
    '''

    tool = 'nextpnr'
    step = chip.get('arg', 'step')
    index = chip.get('arg', 'index')
    _, task = get_tool_task(chip, step, index)

    topmodule = chip.top()

    clobber = False
    chip.set('tool', tool, 'exe', 'nextpnr-ice40')
    chip.set('tool', tool, 'vswitch', '--version')
    chip.set('tool', tool, 'version', '>=0.2', clobber=clobber)

    chip.set('tool', tool, 'task', task, 'option', "", step=step, index=index, clobber=clobber)
    chip.set('tool', tool, 'task', task, 'input', f'{topmodule}.netlist.json',
             step=step, index=index)
    chip.set('tool', tool, 'task', task, 'output', f'{topmodule}.asc', step=step, index=index)




'''
Icepack converts an ASCII bitstream file to a .bin file for the
ICE40 FPGA.

Documentation: https://clifford.at/icestorm

Sources: https://github.com/YosysHQ/icestorm

Installation: https://github.com/YosysHQ/icestorm
'''


#####################################################################
# Make Docs
#####################################################################
def make_docs(chip):
    from siliconcompiler.tools.icepack.bitstream import setup
    setup(chip)
    return chip


################################
#  Custom runtime options
################################
def runtime_options(chip):
    ''' Custom runtime options, returnst list of command line options.
    '''

    topmodule = chip.top()

    cmdlist = []
    cmdlist.append("inputs/" + topmodule + ".asc")
    cmdlist.append("outputs/" + topmodule + ".bit")

    return cmdlist


##################################################
if __name__ == "__main__":

    chip = make_docs()
    chip.write_manifest("icepack.json")



import math


###########################################################################
def calc_area(chip, step=None, index=None):
    '''Calculates the area of a rectilinear diearea.

    Uses the shoelace formulate to calculate the design area using
    the (x,y) point tuples from the 'diearea' parameter. If only diearea
    parameter only contains two points, then the first and second point
    must be the lower left and upper right points of the rectangle.
    (Ref: https://en.wikipedia.org/wiki/Shoelace_formula)

    Args:
        step (str): name of the step to calculate the area from
        index (str): name of the step to calculate the area from

    Returns:
        Design area (float).

    Examples:
        >>> area = chip.calc_area()

    '''

    if not step:
        step = chip.get('arg', 'step')

    if not index:
        index = chip.get('arg', 'index')

    vertices = chip.get('constraint', 'outline', step=step, index=index)

    if len(vertices) == 2:
        width = vertices[1][0] - vertices[0][0]
        height = vertices[1][1] - vertices[0][1]
        area = width * height
    else:
        area = 0.0
        for i in range(len(vertices)):
            j = (i + 1) % len(vertices)
            area += vertices[i][0] * vertices[j][1]
            area -= vertices[j][0] * vertices[i][1]
        area = abs(area) / 2

    return area


###########################################################################
def calc_yield(chip, step=None, index=None, model='poisson'):
    '''Calculates raw die yield.

    Calculates the raw yield of the design as a function of design area
    and d0 defect density. Calculation can be done based on the poisson
    model (default) or the murphy model. The die area and the d0
    parameters are taken from the chip dictionary.

    * Poisson model: dy = exp(-area * d0/100).
    * Murphy model: dy = ((1-exp(-area * d0/100))/(area * d0/100))^2.

    Args:
        step (str): name of the step use for calculation
        index (str): name of the step use for calculation
        model (string): Model to use for calculation (poisson or murphy)

    Returns:
        Design yield percentage (float).

    Examples:
        >>> yield = chip.calc_yield()
        Yield variable gets yield value based on the chip manifest.
    '''

    pdk = chip.get('option', 'pdk')
    d0 = chip.get('pdk', pdk, 'd0')
    if d0 is None:
        chip.error(f"['pdk', {pdk}, 'd0'] has not been set")
    diearea = calc_area(chip, step=step, index=index)

    # diearea is um^2, but d0 looking for cm^2
    diearea = diearea / 10000.0**2

    if model == 'poisson':
        dy = math.exp(-diearea * d0 / 100)
    elif model == 'murphy':
        dy = ((1 - math.exp(-diearea * d0 / 100)) / (diearea * d0 / 100))**2
    else:
        chip.error(f'Unknown yield model: {model}')

    return dy


##########################################################################
def calc_dpw(chip, step=None, index=None):
    '''Calculates dies per wafer.

    Calculates the gross dies per wafer based on the design area, wafersize,
    wafer edge margin, and scribe lines. The calculation is done by starting
    at the center of the wafer and placing as many complete design
    footprints as possible within a legal placement area.

    Args:
        step (str): name of the step use for calculation
        index (str): name of the step use for calculation

    Returns:
        Number of gross dies per wafer (int).

    Examples:
        >>> dpw = chip.calc_dpw()
        Variable dpw gets gross dies per wafer value based on the chip manifest.
    '''

    # PDK information
    pdk = chip.get('option', 'pdk')
    wafersize = chip.get('pdk', pdk, 'wafersize')
    edgemargin = chip.get('pdk', pdk, 'edgemargin')
    hscribe, vscribe = chip.get('pdk', pdk, 'scribe')

    # Design parameters
    diesize = chip.get('constraint', 'outline', step=step, index=index)

    # Convert to mm
    diewidth = (diesize[1][0] - diesize[0][0]) / 1000.0
    dieheight = (diesize[1][1] - diesize[0][1]) / 1000.0

    # Derived parameters
    radius = wafersize / 2 - edgemargin
    stepwidth = diewidth + hscribe
    stepheight = dieheight + vscribe

    # Raster dies out from center until you touch edge margin
    # Work quadrant by quadrant
    dies = 0
    for quad in ('q1', 'q2', 'q3', 'q4'):
        x = 0
        y = 0
        if quad == "q1":
            xincr = stepwidth
            yincr = stepheight
        elif quad == "q2":
            xincr = -stepwidth
            yincr = stepheight
        elif quad == "q3":
            xincr = -stepwidth
            yincr = -stepheight
        elif quad == "q4":
            xincr = stepwidth
            yincr = -stepheight
        # loop through all y values from center
        while math.hypot(0, y) < radius:
            y = y + yincr
            x = xincr
            while math.hypot(x, y) < radius:
                x = x + xincr
                dies = dies + 1

    return dies



#!/usr/bin/env python3
# Copyright 2023 Silicon Compiler Authors. All Rights Reserved.

import argparse
import json
import os
import tempfile
import subprocess
import re

tools = None
data_file = os.path.join(os.path.dirname(__file__), "_tools.json")
with open(data_file, "r") as f:
    tools = json.load(f)


def __make_github_url(url, old_version, new_version):
    if 'github' not in url:
        return None

    if url.endswith('.git'):
        url = url[0:-4]

    return f'{url}/compare/{old_version}...{new_version}'


def bump_commit(tools, tool):
    if "git-url" not in tools[tool]:
        return (None, None)

    if not re.fullmatch(r"[a-f0-9]{40}", tools[tool]["git-commit"]):
        return bump_commit_tag(tools, tool)

    import git

    with tempfile.TemporaryDirectory(prefix=tool) as repo_work_dir:
        repo = git.Repo.clone_from(tools[tool]["git-url"], repo_work_dir)

        return (repo.head.commit.hexsha,
                __make_github_url(tools[tool]["git-url"],
                                  tools[tool]["git-commit"],
                                  repo.head.commit.hexsha))

    return (None, None)


def bump_commit_tag(tools, tool):
    if "git-url" not in tools[tool]:
        return (None, None)

    import git

    version_prefix = 'v'
    if "version-prefix" in tools[tool]:
        version_prefix = tools[tool]["version-prefix"]

    with tempfile.TemporaryDirectory(prefix=tool) as repo_work_dir:
        repo = git.Repo.clone_from(tools[tool]["git-url"], repo_work_dir)

        newest = None
        for tag in repo.tags:
            if not tag.name.startswith(version_prefix):
                continue

            if not newest or tag.commit.committed_datetime > newest.commit.committed_datetime:
                newest = tag
        if newest:
            newest = newest.name
            return (newest,
                    __make_github_url(tools[tool]["git-url"],
                                      tools[tool]["git-commit"],
                                      newest))

    return (None, None)


def bump_version(tools, tool):
    if "git-url" not in tools[tool]:
        return (None, None)

    import git

    with tempfile.TemporaryDirectory(prefix=tool) as repo_work_dir:
        repo = git.Repo.clone_from(tools[tool]["git-url"], repo_work_dir)

        if "run-version" in tools[tool]:
            script = os.path.join(repo_work_dir, 'sc_get_version.sh')
            with open(script, 'w') as f:
                f.write(tools[tool]["run-version"])
            os.chmod(script, 0o700)
            proc = subprocess.run(['bash', script],
                                  stdout=subprocess.PIPE,
                                  stderr=subprocess.STDOUT,
                                  universal_newlines=True,
                                  cwd=repo_work_dir)
            version = proc.stdout.strip()
            if not version:
                return (None, None)

            releasenotes = None
            if "release-notes" in tools[tool]:
                releasenotes = tools[tool]["release-notes"]
            return (version, releasenotes)

        newest = None
        for tag in repo.tags:
            if not newest:
                newest = tag
            else:
                if tag.commit.committed_datetime > newest.commit.committed_datetime:
                    newest = tag
        if newest:
            newest = newest.name
            has_v = False
            if newest[0] == 'v':
                newest = newest[1:]
                has_v = True

            new_version = newest
            old_version = tools[tool]["version"]
            if has_v:
                new_version = f'v{new_version}'
                old_version = f'v{old_version}'

            return (newest,
                    __make_github_url(tools[tool]["git-url"],
                                      old_version,
                                      new_version))

        return (None, None)

    return (None, None)


def has_tool(tool):
    return tool in tools


def get_field(tool, field):
    if field not in tools[tool]:
        return None
    return tools[tool][field]


def get_tools():
    return list(tools.keys())


if __name__ == "__main__":
    supported_tools = ", ".join(get_tools())
    supported_fields = set()
    for tool, fields in tools.items():
        for field in fields:
            supported_fields.add(field)
    supported_fields = ", ".join(supported_fields)

    parser = argparse.ArgumentParser(
        prog="SiliconCompiler Tool Helper",
        description="Maintains current known good versions for all install scripts to use")
    parser.add_argument("--tool", type=str,
                        help=f"Tool name, supported tools: {supported_tools}")
    parser.add_argument("--json_tools", action="store_true",
                        help="Flag to get json matrix used by github to update tools")

    parser.add_argument("--field", type=str,
                        help=f"Field to get information from, supported fields: {supported_fields}")
    parser.add_argument("--bump_commit", action="store_true",
                        help="Flag to indicate that the specified tool should be updated.")

    args = parser.parse_args()

    if args.json_tools:
        json_tools = {'include': []}
        for tool in get_tools():
            field = get_field(tool, "git-url")
            update = get_field(tool, "auto-update")
            if field and update:
                json_tools['include'].append({"tool": tool})
        if len(json_tools['include']) == 0:
            print(json.dumps({}))
        else:
            print(json.dumps(json_tools))
        exit(0)

    if not has_tool(args.tool):
        print(f"{args.tool} is not a supported tool.")
        print(f"Supported tools are: {supported_tools}")
        exit(1)

    if not args.bump_commit:
        tool_fields = tools[args.tool]
        if args.field not in tool_fields:
            print(f"{args.field} is not a supported field for {args.tool}.")
            tool_supported_fields = ", ".join(tool_fields.keys())
            print(f"Supported fields are: {tool_supported_fields}")
            exit(1)

        print(tool_fields[args.field])
        exit(0)

    if "git-commit" in tools[args.tool]:
        new_value, url = bump_commit(tools, args.tool)
        if new_value and tools[args.tool]["git-commit"] != new_value:
            print(f"Updating {args.tool} from {tools[args.tool]['git-commit']} to {new_value}")
            if url:
                print(f'Check {url} for changes')
            tools[args.tool]["git-commit"] = new_value
    elif "version" in tools[args.tool]:
        new_value, url = bump_version(tools, args.tool)
        if new_value and tools[args.tool]["version"] != new_value:
            print(f"Updating {args.tool} from {tools[args.tool]['version']} to {new_value}")
            if url:
                print(f'Check {url} for changes')
            tools[args.tool]["version"] = new_value
    else:
        print('Unsupported update tool')
        exit(1)

    with open(data_file, "w") as f:
        f.write(json.dumps(tools, indent=2))

    exit(0)
